import uuid
import json
import threading
import gevent

from .room import Room

from requests import Session
from werkzeug.http import parse_cookie


class Api(Session):

    def __init__(self, connection):
        super().__init__()

        self.URL_FORMAT = '{}{{}}'.format(connection.hub.adapter.base_url)

        self.headers['cookie'] = connection.ws.environ['HTTP_COOKIE']
        self.headers['host'] = connection.ws.environ['HTTP_HOST']
        self.headers['user-agent'] = connection.ws.environ['HTTP_USER_AGENT']
        FORWARD_IP = connection.hub.app.config.get('FORWARD_IP')
        if FORWARD_IP and FORWARD_IP in connection.ws.environ:
            self.headers['x-forwarded-for'] = connection.ws.environ[FORWARD_IP]

        cookie = parse_cookie(self.headers['cookie'])
        if 'csrftoken' in cookie:
            self.headers['x-csrftoken'] = cookie['csrftoken']

        wz_r = connection.ws.environ.get('werkzeug.request', None)
        if wz_r and 'token' in wz_r.args:
            self.headers['authorization'] = (
                'Token {}'.format(wz_r.args['token'])
            )

    def request(self, method, url, *args, **kwargs):
        url = self.URL_FORMAT.format(url)
        return super().request(method, url, *args, **kwargs)


class Connection():
    ws = None
    hub = None
    user_id = None

    def __init__(self, hub, ws):
        self.ws = ws
        self.hub = hub
        self.subscriptions = {}
        self.allowed_rooms = []
        self.uuid = uuid.uuid4()

        # Nasty hack
        try:
            ws.connection = self
        except AttributeError:
            pass

        self.api = Api(self)

    def get_write_lock(self):
        if not hasattr(self, '_write_lock'):
            self._write_lock = threading.Lock()
        return self._write_lock


    def handle_auth_success(self, data):
        self.user_id = data
        for key in self.hub.app.config.get('USER_ID_PATH', ['user', 'id']):
            try:
                self.user_id = self.user_id[key]
            except (TypeError, KeyError, IndexError):
                self.user_id = None
                break

        self.allowed_rooms = data.get('allowed_rooms', [])

        self.send({'allowed_rooms': self.allowed_rooms})

    def handle(self, message):
        if message == 'ping':
            self.send_raw('pong')
            for hook in self.hub.ping_hooks:
                hook(self)
            return

        m = json.loads(message)

        requestId = m.get('requestId', None)

        if m['type'] == 'subscribe':
            return self.handle_subscribe(m)

        if m['type'] == 'unsubscribe':
            return self.handle_unsubscribe(m)

        self.send({
            'requestId': requestId,
            'code': 'error',
            'message': 'message-type-not-allowed',
        })

    # Check that the requested room's keys match any of the allowed rooms,
    def is_room_allowed(self, requested_room_dict):
        def room_matches(ar):
            if len(requested_room_dict.keys()) != len(ar.keys()):
                return False
            for allowed_key, allowed_value in ar.items():
                if allowed_key not in requested_room_dict:
                    return False
                if allowed_value == '*':
                    continue
                if requested_room_dict[allowed_key] != allowed_value:
                    return False

            return True

        for ar in self.allowed_rooms:
            if room_matches(ar):
                return True

        return False


    def handle_subscribe(self, m):
        room_dict = m.get('room', None)
        room_hash = Room.hash_dict(room_dict)

        if not self.is_room_allowed(room_dict):
            self.send({
                'requestId': m['requestId'],
                'code': 'error',
                'message': 'room-not-found',
            })
            return

        sub = self.hub.subscribe(self, m, room_hash)
        self.subscriptions[m['requestId']] = sub
        self.send({
            'requestId': m['requestId'],
            'code': 'success',
        })

    def handle_unsubscribe(self, m):
        reqId = m['requestId']
        if reqId not in self.subscriptions:
            self.send({
                'requestId': m['requestId'],
                'code': 'error',
                'message': 'not-subscribed',
            })
            return

        sub = self.subscriptions[reqId]
        sub.stop()

        self.subscriptions.pop(reqId)
        self.send({
            'requestId': reqId,
            'code': 'success',
        })

    def unsubscribe_all(self):
        for room in [sub.room for sub in self.subscriptions.values()]:
            room.remove_connection(self)

        self.subscriptions = {}

    def send_raw(self, message):
        if self.ws.closed:
            return
        def _send(ws, message):
            ws.stream.handler.socket.settimeout(0.1)
            ws.send(message)
            ws.stream.handler.socket.settimeout(None)

        gevent.spawn(_send, self.ws, message)

    def send(self, message):
        self.send_raw(json.dumps(message))



from flask import make_response
from .room import Room
from .connection import Connection


class Adapter:
    '''
    Translates incoming requests from the django app to hub actions
    and vice versa.
    '''

    def __init__(self, app):
        self.app = app
        self.base_url = app.config['API_URL']

    def check_auth(self, connection):
        res = connection.api.get('bootstrap/')

        if res.status_code != 200:
            return False

        connection.handle_auth_success(res.json())
        return True


class Hub:
    '''
    Manages all connections.
    '''

    def __init__(self, app):
        self.app = app
        self.rooms = {}
        self.adapter = Adapter(app)

        self.connect_hooks = []
        self.ping_hooks = []
        self.disconnect_hooks = []

    def handle_trigger(self, body):
        if 'rooms' not in body:
            return make_response('rooms not specified', 400)
        if 'data' not in body:
            return make_response('data not specified', 400)

        closed_connections = []
        for room_dict in body['rooms']:
            room_hash = Room.hash_dict(room_dict)
            if room_hash not in self.rooms:
                continue
            connections = self.rooms[room_hash].publish(body['data'])
            closed_connections.extend(connections)

        for r_name, room in self.rooms.copy().items():
            for c in closed_connections:
                c.unsubscribe_all()

        return make_response('publish success')

    def add_if_auth(self, ws):
        connection = Connection(self, ws)

        auth = self.adapter.check_auth(connection)
        if not auth:
            ws.close()
            return False

        return connection

    def get_or_create_room(self, room_hash):
        if room_hash in self.rooms:
            return self.rooms[room_hash]

        room = Room(room_hash, self)
        self.rooms[room_hash] = room
        return room

    def subscribe(self, connection, request, room_hash):
        room = self.get_or_create_room(room_hash)
        sub = room.subscribe(connection, request)
        return sub

    def close_room(self, room):
        # Clone self.rooms because they may be removed when closed
        rooms = self.rooms.copy()
        for r_name, r in rooms.items():
            if r == room:
                del self.rooms[r_name]

    def on_connect(self, func):
        self.connect_hooks.append(func)
        return func

    def on_ping(self, func):
        self.ping_hooks.append(func)
        return func

    def on_disconnect(self, func):
        self.disconnect_hooks.append(func)
        return func



from flask_sockets import Sockets
from flask import Flask, request
import logging
import json
import gevent

def create_app(settings=None):
    app = Flask(__name__)

    if settings:
        app.config.from_object(settings)

    sockets = Sockets(app)

    from .hub import Hub
    app.hub = Hub(app)

    @sockets.route('/ws/')
    def open_socket(ws):
        connection = app.hub.add_if_auth(ws)
        if not connection:
            # todo manage connection close
            return

        for hook in app.hub.connect_hooks:
            hook(connection)

        def process(connection, message):
            if message:
                try:
                    connection.handle(message)
                except Exception as e:
                    logging.error(e, exc_info=True)


        while not connection.ws.closed:
            message = connection.ws.receive()
            gevent.spawn(process, connection, message)

        for hook in app.hub.disconnect_hooks:
            hook(connection)

    @app.route('/trigger/', methods=['POST'])
    def handle_trigger():
        data = json.loads(request.data.decode())
        return app.hub.handle_trigger(data)

    return app
